
// example of floating point math expression error checking
// link with math library (gcc -lm)

#include <stdio.h>              // printf()
#include <string.h>             // strerror()

#include <math.h>
#include <errno.h>
#include <fenv.h>

/*
    From manual:

    When an error occurs, most library functions indicate this fact by
    returning a special value (e.g., -1 or NULL).  Because they typically return
    a floating-point number, the mathematical functions declared in <math.h>
    indicate an  error  using  other mechanisms.  There are two error-reporting
    mechanisms: the older one sets errno; the newer one uses the floating-point
    exception mechanism (the use of feclearexcept(3) and fetestexcept(3), as
    outlined below) described in fenv(3).

    A  portable  program that needs to check for an error from a mathematical
    function should set errno to zero, and make the following call

       feclearexcept(FE_ALL_EXCEPT);

    before calling a mathematical function.

    Upon return from the mathematical function, if errno is nonzero, or the
    following call (see fenv(3)) returns nonzero

       fetestexcept(FE_INVALID | FE_DIVBYZERO | FE_OVERFLOW | FE_UNDERFLOW);

    then an error occurred in the mathematical function.

    ...

    The gcc -fno-math-errno option causes the executable to employ
    implementations of some mathematical functions that are faster than the
    standard implementations, but do not set errno on error.
    (The gcc -ffast-math option also enables -fno-math-errno.)
    An error can still be tested for using fetestexcept(3).

*/

double test_log(double x)
{
    if (isnan(x) || islessequal(x, 0)) {
        // possible detection of bad input values
        // fprintf(stderr,"TEST: x is NaN or x<=0\n");
        // return ?
    }
    // prepare for error checking:
    errno = 0;
    feclearexcept(FE_ALL_EXCEPT);
    // compute value:
    double r = log(x);
    // check possible errors:
    char *msg = strerror(errno);
    int flags = fetestexcept(
                  FE_INVALID | FE_DIVBYZERO |
                  FE_OVERFLOW | FE_UNDERFLOW );
    printf("log(%g) == %g  \t flags=0x%x  errno=%d \n",
                 x,     r,            flags,    errno);
    if (errno)
        printf("  errno: %d \"%s\"\n", errno, msg);
    if (flags) {
        printf("  flags: {");
        if (flags & FE_INVALID)
            printf(" FE_INVALID");
        if (flags & FE_DIVBYZERO)
            printf(" FE_DIVBYZERO");
        if (flags & FE_OVERFLOW)
            printf(" FE_OVERFLOW");
        if (flags & FE_UNDERFLOW)
            printf(" FE_UNDERFLOW");
        printf(" } \n");
    }
    return r;
}

int main()
{
    test_log(NAN);
    test_log(-1.0);
    test_log(0.0);
    test_log(0.1);
    test_log(2.0);
    test_log(10.0);
}
